@@ -187,11 +187,23 @@ be set to $(FilePath).
 void SciTEBase::ReadDirectoryPropFile() {
 	propsDirectory.Clear();
 
-	if (props.GetInt("properties.directory.enable") != 0) {
-		FilePath propfile = GetDirectoryPropertiesFileName();
-		props.Set("SciteDirectoryHome", propfile.Directory().AsUTF8().c_str());
+	propsDirectory.SetPath("FilePath", filePath);
+	propsDirectory.SetPath("FileDir", filePath.Directory());
+	propsDirectory.SetPath("FileName", filePath.BaseName());
+	propsDirectory.SetPath("FileExt", filePath.Extension());
+	propsDirectory.SetPath("FileNameExt", FileNameExt());
 
-		propsDirectory.Read(propfile, propfile.Directory(), filter, nullptr, 0);
+	if (props.GetInt("properties.directory.enable") != 0) {
+		const FilePath propfile = GetDirectoryPropertiesFileName();
+		const FilePath propfileDirectory = propfile.Directory();
+		const GUI::gui_string relPath = propfileDirectory.RelativePathTo(filePath);
+		propsDirectory.Set("RelativePath", GUI::UTF8FromString(relPath));
+
+		props.SetPath("SciteDirectoryHome", propfileDirectory);
+
+		propsDirectory.Read(propfile, propfileDirectory, filter, nullptr, 0);
+	} else {
+		propsDirectory.SetPath("RelativePath", filePath.Name());
 	}
 }
 
@@ -262,7 +274,7 @@ const char *SciTEBase::GetNextPropItem(
 	const char *pStart,	/**< the property string to parse for the first call,
 						 * pointer returned by the previous call for the following. */
 	char *pPropItem,	///< pointer on a buffer receiving the requested prop item
-	int maxLen) {		///< size of the above buffer
+	size_t maxLen) noexcept {	///< size of the above buffer
 	ptrdiff_t size = maxLen - 1;
 
 	*pPropItem = '\0';
@@ -562,6 +574,7 @@ static const char *propertiesToForward[] = {
 	"lexer.json.escape.sequence",
 	"lexer.julia.highlight.lexerror",
 	"lexer.julia.highlight.typeannotation",
+	"lexer.markdown.header.eolfill",
 	"lexer.metapost.comment.process",
 	"lexer.metapost.interface.default",
 	"lexer.nim.raw.strings.highlight.ident",
@@ -844,10 +857,8 @@ void SciTEBase::ReadProperties() {
 		}
 	}
 
-	FilePath homepath = GetSciteDefaultHome();
-	props.Set("SciteDefaultHome", homepath.AsUTF8().c_str());
-	homepath = GetSciteUserHome();
-	props.Set("SciteUserHome", homepath.AsUTF8().c_str());
+	props.SetPath("SciteDefaultHome", GetSciteDefaultHome());
+	props.SetPath("SciteUserHome", GetSciteUserHome());
 
 	for (size_t i=0; propertiesToForward[i]; i++) {
 		ForwardPropertyToEditor(propertiesToForward[i]);
@@ -859,7 +870,7 @@ void SciTEBase::ReadProperties() {
 		apisFileNames = props.GetNewExpandString("api.", fileNameForExtension.c_str());
 	}
 
-	props.Set("APIPath", apisFileNames.c_str());
+	props.Set("APIPath", apisFileNames);
 
 	FilePath fileAbbrev = GUI::StringFromUTF8(props.GetNewExpandString("abbreviations.", fileNameForExtension.c_str()));
 	if (!fileAbbrev.IsSet())
@@ -869,7 +880,7 @@ void SciTEBase::ReadProperties() {
 		ReadAbbrevPropFile();
 	}
 
-	props.Set("AbbrevPath", pathAbbreviations.AsUTF8().c_str());
+	props.SetPath("AbbrevPath", pathAbbreviations);
 
 	const SA::Technology tech = static_cast<SA::Technology>(props.GetInt("technology"));
 	wEditor.SetTechnology(tech);
@@ -1860,10 +1871,8 @@ void SciTEBase::ReadPropertiesInitial() {
 	}
 	// end load the user defined short cut props
 
-	FilePath homepath = GetSciteDefaultHome();
-	props.Set("SciteDefaultHome", homepath.AsUTF8().c_str());
-	homepath = GetSciteUserHome();
-	props.Set("SciteUserHome", homepath.AsUTF8().c_str());
+	props.SetPath("SciteDefaultHome", GetSciteDefaultHome());
+	props.SetPath("SciteUserHome", GetSciteUserHome());
 }
 
 FilePath SciTEBase::GetDefaultPropertiesFileName() {
