@@ -29,8 +29,232 @@
 
 namespace {
 
+constexpr char MakeLowerCase(char c) noexcept {
+	if (c >= 'A' && c <= 'Z') {
+		return c - 'A' + 'a';
+	} else {
+		return c;
+	}
+}
+
+[[maybe_unused]] void LowerCaseAZ(std::string &s) {
+	std::transform(s.begin(), s.end(), s.begin(), MakeLowerCase);
+}
+
+int IntFromString(std::u32string_view s) noexcept {
+	if (s.empty()) {
+		return 0;
+	}
+	const bool negate = s.front() == '-';
+	if (negate) {
+		s.remove_prefix(1);
+	}
+	int value = 0;
+	while (!s.empty()) {
+		value = value * 10 + s.front() - '0';
+		s.remove_prefix(1);
+	}
+	return negate ? -value : value;
+}
+
+bool PatternMatch(std::u32string_view pattern, std::u32string_view text) noexcept {
+	if (pattern == text) {
+		return true;
+	} else if (pattern.empty()) {
+		return false;
+	} else if (pattern.front() == '\\') {
+		pattern.remove_prefix(1);
+		if (pattern.empty()) {
+			// Escape with nothing being escaped
+			return false;
+		}
+		if (text.empty()) {
+			return false;
+		}
+		if (pattern.front() == text.front()) {
+			pattern.remove_prefix(1);
+			text.remove_prefix(1);
+			return PatternMatch(pattern, text);
+		}
+		return false;
+	} else if (pattern.front() == '*') {
+		pattern.remove_prefix(1);
+		if (!pattern.empty() && pattern.front() == '*') {
+			pattern.remove_prefix(1);
+			// "**" matches anything including "/"
+			while (!text.empty()) {
+				if (PatternMatch(pattern, text)) {
+					return true;
+				}
+				text.remove_prefix(1);
+			}
+		} else {
+			while (!text.empty()) {
+				if (PatternMatch(pattern, text)) {
+					return true;
+				}
+				if (text.front() == '/') {
+					// "/" not matched by single "*"
+					return false;
+				}
+				text.remove_prefix(1);
+			}
+		}
+		assert(text.empty());
+		// Consumed whole text with wildcard so match if pattern consumed
+		return pattern.empty();
+	} else if (text.empty()) {
+		return false;
+	} else if (pattern.front() == '?') {
+		if (text.front() == '/') {
+			return false;
+		}
+		pattern.remove_prefix(1);
+		text.remove_prefix(1);
+		return PatternMatch(pattern, text);
+	} else if (pattern.front() == '[') {
+		pattern.remove_prefix(1);
+		if (pattern.empty()) {
+			return false;
+		}
+		const bool positive = pattern.front() != '!';
+		if (!positive) {
+			pattern.remove_prefix(1);
+			if (pattern.empty()) {
+				return false;
+			}
+		}
+		bool inSet = false;
+		if (!pattern.empty() && pattern.front() == ']') {
+			// First is allowed to be ']'
+			if (pattern.front() == text.front()) {
+				inSet = true;
+			}
+			pattern.remove_prefix(1);
+		}
+		char32_t start = 0;
+		while (!pattern.empty() && pattern.front() != ']') {
+			if (pattern.front() == '-') {
+				pattern.remove_prefix(1);
+				if (!pattern.empty()) {
+					const char32_t end = pattern.front();
+					if ((text.front() >= start) && (text.front() <= end)) {
+						inSet = true;
+					}
+				}
+			} else if (pattern.front() == text.front()) {
+				inSet = true;
+			}
+			if (!pattern.empty()) {
+				start = pattern.front();
+				pattern.remove_prefix(1);
+			}
+		}
+		if (!pattern.empty()) {
+			pattern.remove_prefix(1);
+		}
+		if (inSet != positive) {
+			return false;
+		}
+		text.remove_prefix(1);
+		return PatternMatch(pattern, text);
+	} else if (pattern.front() == '{') {
+		if (pattern.length() < 2) {
+			return false;
+		}
+		const size_t endParen = pattern.find('}');
+		if (endParen == std::u32string_view::npos) {
+			// Malformed {x} pattern
+			return false;
+		}
+		std::u32string_view parenExpression = pattern.substr(1, endParen - 1);
+		bool inSet = false;
+		const size_t dotdot = parenExpression.find(U"..");
+		if (dotdot != std::u32string_view::npos) {
+			// Numeric range: {10..20}
+			const std::u32string_view firstRange = parenExpression.substr(0, dotdot);
+			const std::u32string_view lastRange = parenExpression.substr(dotdot+2);
+			if (firstRange.empty() || lastRange.empty()) {
+				// Malformed {s..e} range pattern
+				return false;
+			}
+			const size_t endInteger = text.find_last_of(U"-0123456789");
+			if (endInteger == std::u32string_view::npos) {
+				// No integer in text
+				return false;
+			}
+			const std::u32string_view intPart = text.substr(0, endInteger+1);
+			const int first = IntFromString(firstRange);
+			const int last = IntFromString(lastRange);
+			const int value = IntFromString(intPart);
+			if ((value >= first) && (value <= last)) {
+				inSet = true;
+				text.remove_prefix(intPart.length());
+			}
+		} else {
+			// Alternates: {a,b,cd}
+			size_t comma = parenExpression.find(',');
+			for (;;) {
+				const bool finalAlt = comma == std::u32string_view::npos;
+				const std::u32string_view oneAlt = finalAlt ? parenExpression :
+					parenExpression.substr(0, comma);
+				if (oneAlt == text.substr(0, oneAlt.length())) {
+					// match
+					inSet = true;
+					text.remove_prefix(oneAlt.length());
+					break;
+				}
+				if (finalAlt) {
+					break;
+				}
+				parenExpression.remove_prefix(oneAlt.length() + 1);
+				comma = parenExpression.find(',');
+			}
+		}
+		if (!inSet) {
+			return false;
+		}
+		pattern.remove_prefix(endParen + 1);
+		return PatternMatch(pattern, text);
+	} else if (pattern.front() == text.front()) {
+		pattern.remove_prefix(1);
+		text.remove_prefix(1);
+		return PatternMatch(pattern, text);
+	}
+	return false;
+}
+
+bool PathMatch(std::string pattern, std::string relPath) {
+#if defined(_WIN32)
+	// Convert Windows path separators to Unix
+	std::replace(relPath.begin(), relPath.end(), '\\', '/');
+#endif
+#if defined(_WIN32) || defined(__APPLE__)
+	// Case-insensitive, only does ASCII but fine for test example files
+	LowerCaseAZ(pattern);
+	LowerCaseAZ(relPath);
+#endif
+	const std::u32string patternU32 = UTF32FromUTF8(pattern);
+	const std::u32string relPathU32 = UTF32FromUTF8(relPath);
+	if (PatternMatch(patternU32, relPathU32)) {
+		return true;
+	}
+	const size_t lastSlash = relPathU32.rfind('/');
+	if (lastSlash == std::string::npos) {
+		return false;
+	}
+	// Match against just filename
+	const std::u32string fileNameU32 = relPathU32.substr(lastSlash+1);
+	return PatternMatch(patternU32, fileNameU32);
+}
+
 constexpr std::string_view suffixStyled = ".styled";
 constexpr std::string_view suffixFolded = ".folded";
+constexpr std::string_view lexerPrefix = "lexer.*";
+constexpr std::string_view prefixIf = "if ";
+constexpr std::string_view prefixMatch = "match ";
+constexpr std::string_view prefixEqual = "= ";
+constexpr std::string_view prefixComment = "#";
 
 std::string ReadFile(std::filesystem::path path) {
 	std::ifstream ifs(path, std::ios::binary);
@@ -40,6 +264,7 @@ std::string ReadFile(std::filesystem::path path) {
 }
 
 std::string MarkedDocument(const Scintilla::IDocument *pdoc) {
+	assert(pdoc);
 	std::ostringstream os(std::ios::binary);
 	char prevStyle = -1;
 	for (Sci_Position pos = 0; pos < pdoc->Length(); pos++) {
@@ -71,6 +296,7 @@ void PrintLevel(std::ostringstream &os, int level) {
 }
 
 std::string FoldedDocument(const Scintilla::IDocument *pdoc) {
+	assert(pdoc);
 	std::ostringstream os(std::ios::binary);
 	Sci_Position linePrev = -1;
 	char ch = '\0';
@@ -94,12 +320,136 @@ std::pair<std::string, std::string> MarkedAndFoldedDocument(const Scintilla::IDo
 	return { MarkedDocument(pdoc), FoldedDocument(pdoc) };
 }
 
+std::vector<std::string> StringSplit(const std::string_view &text, int separator) {
+	std::vector<std::string> vs(text.empty() ? 0 : 1);
+	for (std::string_view::const_iterator it = text.begin(); it != text.end(); ++it) {
+		if (*it == separator) {
+			vs.push_back(std::string());
+		} else {
+			vs.back() += *it;
+		}
+	}
+	return vs;
+}
+
+static constexpr bool IsSpaceOrTab(char ch) noexcept {
+	return (ch == ' ') || (ch == '\t');
+}
+
 class PropertyMap {
+
+	std::string Evaluate(std::string_view text) {
+		if (text.find(' ') != std::string_view::npos) {
+			if (text.starts_with(prefixEqual)) {
+				const std::string_view sExpressions = text.substr(prefixEqual.length());
+				std::vector<std::string> parts = StringSplit(sExpressions, ';');
+				if (parts.size() > 1) {
+					for (size_t part = 1; part < parts.size(); part++) {
+						if (parts.at(part) != parts.at(0)) {
+							return "0";
+						}
+					}
+					return "1";
+				}
+			}
+			return {};
+		} else {
+			std::optional<std::string> value = GetProperty(text);
+			if (value) {
+				return *value;
+			}
+			return {};
+		}
+	}
+
+	std::string Expand(std::string withVars) {
+		constexpr size_t maxVars = 100;
+		size_t varStart = withVars.rfind("$(");
+		for (size_t count = 0; (count < maxVars) && (varStart != std::string::npos); count++) {
+			const size_t varEnd = withVars.find(')', varStart + 2);
+			if (varEnd == std::string::npos) {
+				break;
+			}
+
+			const std::string_view whole = withVars;
+			const std::string_view var = whole.substr(varStart + 2, varEnd - (varStart + 2));
+			const std::string val = Evaluate(var);
+
+			withVars.erase(varStart, varEnd - varStart + 1);
+			withVars.insert(varStart, val);
+
+			varStart = withVars.rfind("$(");
+		}
+		return withVars;
+	}
+
+	std::vector<std::string> GetFilePatterns(const std::string &key) const {
+		std::vector<std::string> exts;
+		// Malformed patterns are skipped if we require the whole prefix here;
+		// a fuzzy search lets us collect and report them
+		const size_t patternStart = key.find('*');
+		if (patternStart == std::string::npos)
+			return exts;
+
+		const std::string patterns = key.substr(patternStart);
+		for (const std::string &pat : StringSplit(patterns, ';')) {
+			// Only accept patterns in the form *.xyz
+			if (pat.starts_with("*.") && pat.length() > 2) {
+				exts.push_back(pat.substr(1));
+			} else {
+				std::cout << "\n"
+					  << "Ignoring bad file pattern '" << pat << "' in list " << patterns << "\n";
+			}
+		}
+		return exts;
+	}
+
+	bool ProcessLine(std::string_view text, bool ifIsTrue) {
+		// If clause ends with first non-indented line
+		if (!ifIsTrue && (text.empty() || IsSpaceOrTab(text.at(0)))) {
+			return false;
+		}
+		ifIsTrue = true;
+		if (text.starts_with(prefixIf)) {
+			const std::string value = Expand(std::string(text.substr(prefixIf.length())));
+			if (value == "0" || value == "") {
+				ifIsTrue = false;
+			}
+		} else if (text.starts_with(prefixMatch)) {
+			std::optional<std::string> fileNameExt = GetProperty("FileNameExt");
+			if (fileNameExt) {
+				std::string pattern(text.substr(prefixMatch.length()));
+				// Remove trailing white space
+				while (!pattern.empty() && IsSpaceOrTab(pattern.back())) {
+					pattern.pop_back();
+				}
+				ifIsTrue = PathMatch(pattern, *fileNameExt);
+			} else {
+				ifIsTrue = false;
+			}
+		} else {
+			while (!text.empty() && IsSpaceOrTab(text.at(0))) {
+				text.remove_prefix(1);
+			}
+			if (text.starts_with(prefixComment)) {
+				return ifIsTrue;
+			}
+			const size_t positionEquals = text.find("=");
+			if (positionEquals != std::string::npos) {
+				const std::string key(text.substr(0, positionEquals));
+				const std::string_view value = text.substr(positionEquals + 1);
+				properties[key] = value;
+			}
+		}
+		return ifIsTrue;
+	}
+
 public:
 	using PropMap = std::map<std::string, std::string>;
 	PropMap properties;
 
 	void ReadFromFile(std::filesystem::path path) {
+		bool ifIsTrue = true;
 		std::ifstream ifs(path);
 		std::string line;
 		std::string logicalLine;
@@ -112,12 +462,7 @@ public:
 			if (logicalLine.ends_with("\\")) {
 				logicalLine.pop_back();
 			} else {
-				const size_t positionEquals = logicalLine.find("=");
-				if (positionEquals != std::string::npos) {
-					const std::string key = logicalLine.substr(0, positionEquals);
-					const std::string value = logicalLine.substr(positionEquals + 1);
-					properties[key] = value;
-				}
+				ifIsTrue = ProcessLine(logicalLine, ifIsTrue);
 				logicalLine.clear();
 			}
 		}
@@ -137,6 +482,15 @@ public:
 				const std::string keySuffix = key.substr(keyPrefix.length());
 				if (fileName.ends_with(keySuffix)) {
 					return val;
+				} else if (key.find(';') != std::string::npos) {
+					// It may be the case that a suite of test files with various extensions are
+					// meant to share a common configuration, so try to find a matching
+					// extension in a delimited list, e.g., lexer.*.html;*.php;*.asp=hypertext
+					for (const std::string &ext : GetFilePatterns(key)) {
+						if (fileName.ends_with(ext)) {
+							return val;
+						}
+					}
 				}
 			}
 		}
@@ -171,7 +525,8 @@ bool CheckSame(std::string_view augmentedText, std::string_view augmentedTextNew
 	}
 	const size_t lineNumber = FirstLineDifferent(augmentedText, augmentedTextNew) + 1;
 	std::cout << "\n" << path.string() << ":" << lineNumber << ":";
-	std::cout << " has different " << item << "\n\n";
+	const std::string differenceType = augmentedText.empty() ? "new" : "different";
+	std::cout << " has " << differenceType << " " << item << "\n\n";
 	std::filesystem::path pathNew = path;
 	pathNew += suffix;
 	pathNew += ".new";
@@ -204,6 +559,7 @@ int UnixToWindows(std::string &s) {
 const std::string BOM = "\xEF\xBB\xBF";
 
 void StyleLineByLine(TestDocument &doc, Scintilla::ILexer5 *plex) {
+	assert(plex);
 	Scintilla::IDocument *pdoc = &doc;
 	const Sci_Position lines = doc.LineFromPosition(doc.Length());
 	Sci_Position startLine = 0;
@@ -218,7 +574,8 @@ void StyleLineByLine(TestDocument &doc, Scintilla::ILexer5 *plex) {
 	}
 }
 
-void TestCRLF(std::filesystem::path path, const std::string s, Scintilla::ILexer5 *plex, bool disablePerLineTests) {
+bool TestCRLF(std::filesystem::path path, const std::string s, Scintilla::ILexer5 *plex, bool disablePerLineTests) {
+	bool success = true;
 	// Convert all line ends to \r\n to check if styles change between \r and \n which makes
 	// it difficult to test on different platforms when files may have line ends changed.
 	std::string text = s;
@@ -244,6 +601,7 @@ void TestCRLF(std::filesystem::path path, const std::string s, Scintilla::ILexer
 				std::cout << path.string() << ":" << line << ":" <<
 					" different styles between \\r and \\n at " <<
 					pos << ": " << prevStyle << ", " << styleNow << "\n";
+				success = false;
 			}
 			line++;
 		}
@@ -265,9 +623,11 @@ void TestCRLF(std::filesystem::path path, const std::string s, Scintilla::ILexer
 
 	if (styledText != styledTextUnix) {
 		std::cout << "\n" << path.string() << ":1: has different styles with \\n versus \\r\\n line ends\n\n";
+		success = false;
 	}
 	if (foldedText != foldedTextUnix) {
 		std::cout << "\n" << path.string() << ":1: has different folds with \\n versus \\r\\n line ends\n\n";
+		success = false;
 	}
 
 	// Test line by line lexing/folding with Unix \n line ends
@@ -277,14 +637,21 @@ void TestCRLF(std::filesystem::path path, const std::string s, Scintilla::ILexer
 		// Convert results from \n to \r\n run
 		UnixToWindows(styledTextNewPerLine);
 		UnixToWindows(foldedTextNewPerLine);
-		CheckSame(styledTextUnix, styledTextNewPerLine, "per-line styles \\n", suffixStyled, path);
-		CheckSame(foldedTextUnix, foldedTextNewPerLine, "per-line folds \\n", suffixFolded, path);
+		if (!CheckSame(styledTextUnix, styledTextNewPerLine, "per-line styles \\n", suffixStyled, path)) {
+			success = false;
+		}
+		if (!CheckSame(foldedTextUnix, foldedTextNewPerLine, "per-line folds \\n", suffixFolded, path)) {
+			success = false;
+		}
 	}
 
 	plex->Release();
+	return success;
 }
 
 void TestILexer(Scintilla::ILexer5 *plex) {
+	assert(plex);
+
 	// Test each method of the ILexer interface.
 	// Mostly ensures there are no crashes when calling methods.
 	// Some methods are tested later (Release, Lex, Fold).
@@ -320,11 +687,11 @@ void TestILexer(Scintilla::ILexer5 *plex) {
 	[[maybe_unused]] const int lineEndTypes = plex->LineEndTypesSupported();
 	assert(lineEndTypes == 0 || lineEndTypes == 1);
 
-	if (const char *bases = plex->GetSubStyleBases()) {
+	if (std::string_view bases = plex->GetSubStyleBases(); !bases.empty()) {
 		// Allocate a substyle for each possible style
-		while (*bases) {
+		while (!bases.empty()) {
 			constexpr int newStyles = 3;
-			const int base = *bases;
+			const int base = bases.front();
 			const int baseStyle = plex->AllocateSubStyles(base, newStyles);
 			[[maybe_unused]] const int styleBack = plex->StyleFromSubStyle(baseStyle);
 			assert(styleBack == base);
@@ -333,7 +700,7 @@ void TestILexer(Scintilla::ILexer5 *plex) {
 			assert(start == baseStyle);
 			[[maybe_unused]] const int len = plex->SubStylesLength(base);
 			assert(len == newStyles);
-			bases++;
+			bases.remove_prefix(1);
 		}
 		plex->FreeSubStyles();
 	}
@@ -359,11 +726,13 @@ void TestILexer(Scintilla::ILexer5 *plex) {
 }
 
 void SetProperties(Scintilla::ILexer5 *plex, const PropertyMap &propertyMap, std::string_view fileName) {
+	assert(plex);
+
 	// Set keywords, keywords2, ... keywords9, for this file
 	for (int kw = 0; kw < 10; kw++) {
 		std::string kwChoice("keywords");
 		if (kw > 0) {
-			kwChoice.push_back('1' + kw);
+			kwChoice.push_back(static_cast<char>('1' + kw));
 		}
 		kwChoice.append(".*");
 		std::optional<std::string> keywordN = propertyMap.GetPropertyForFile(kwChoice, fileName);
@@ -374,9 +743,7 @@ void SetProperties(Scintilla::ILexer5 *plex, const PropertyMap &propertyMap, std
 
 	// Set parameters of lexer
 	for (auto const &[key, val] : propertyMap.properties) {
-		if (key.starts_with("#")) {
-			// Ignore comments
-		} else if (key.starts_with("lexer.*")) {
+		if (key.starts_with("lexer.*")) {
 			// Ignore as processed earlier
 		} else if (key.starts_with("keywords")) {
 			// Ignore as processed earlier
@@ -386,7 +753,6 @@ void SetProperties(Scintilla::ILexer5 *plex, const PropertyMap &propertyMap, std
 	}
 }
 
-const char *lexerPrefix = "lexer.*";
 
 bool TestFile(const std::filesystem::path &path, const PropertyMap &propertyMap) {
 	// Find and create correct lexer
@@ -455,16 +821,16 @@ bool TestFile(const std::filesystem::path &path, const PropertyMap &propertyMap)
 
 	plex->Release();
 
-	Scintilla::ILexer5 *plexCRLF = Lexilla::MakeLexer(*language);
-	SetProperties(plexCRLF, propertyMap, path.filename().string());
-	TestCRLF(path, text, plexCRLF, disablePerLineTests);
+	if (success) {
+		Scintilla::ILexer5 *plexCRLF = Lexilla::MakeLexer(*language);
+		SetProperties(plexCRLF, propertyMap, path.filename().string());
+		success = TestCRLF(path, text, plexCRLF, disablePerLineTests);
+	}
 
 	return success;
 }
 
 bool TestDirectory(std::filesystem::path directory, std::filesystem::path basePath) {
-	PropertyMap properties;
-	properties.ReadFromFile(directory / "SciTE.properties");
 	bool success = true;
 	for (auto &p : std::filesystem::directory_iterator(directory)) {
 		if (!p.is_directory()) {
@@ -473,6 +839,9 @@ bool TestDirectory(std::filesystem::path directory, std::filesystem::path basePa
 				extension != suffixFolded) {
 				const std::filesystem::path relativePath = p.path().lexically_relative(basePath);
 				std::cout << "Lexing " << relativePath.string() << '\n';
+				PropertyMap properties;
+				properties.properties["FileNameExt"] = p.path().filename().string();
+				properties.ReadFromFile(directory / "SciTE.properties");
 				if (!TestFile(p, properties)) {
 					success = false;
 				}
